use std::{fs, io, path::Path};
use thiserror::Error;

use crate::{
    cli::{CliSolutionToSolve, GivenDay, GivenTask},
    solutions::ALL_SOLUTIONS,
    AppResult,
};

pub fn solve_given(given_day: GivenDay, given_task: GivenTask, content: &str) -> AppResult<String> {
    let found_task = {
        let day: u32 = given_day.into();
        let task: u32 = given_task.into();
        let found_day = ALL_SOLUTIONS
            .get(day.saturating_sub(1) as usize)
            .ok_or_else(|| CouldNotSolveError::DayNotFound(day))?;
        found_day
            .get(task.saturating_sub(1) as usize)
            .ok_or_else(|| CouldNotSolveError::TaskNotFound { day, task })
    }?;

    let solved = (found_task)(content);
    Ok(solved)
}

pub fn solve_given_from_cli(args: &CliSolutionToSolve) -> AppResult<String> {
    let puzzel_input = try_read_from_file_if_demanded(args)?;
    let output = solve_given(args.day(), args.task(), &puzzel_input)?;
    Ok(output)
}

fn try_read_from_file_if_demanded(args: &CliSolutionToSolve) -> io::Result<String> {
    let content = if args.read_as_file() {
        let path = Path::new(args.input());
        fs::read_to_string(path)?
    } else {
        args.input().to_string()
    };
    Ok(content)
}

#[derive(Debug, Error)]
enum CouldNotSolveError {
    #[error("There is no solution for the day {0}")]
    DayNotFound(u32),
    #[error("There is not solution for task {task} under the day {day}")]
    TaskNotFound { day: u32, task: u32 },
    #[error("Could not read puzzel input from the given file\n {0}")]
    CouldNotReadFromFile(#[from] io::Error),
}
